/***************************************************************************
 *            exif.h
 *
 *  Sun Apr  8 23:58:46 2007
 *  Copyright  2007  Veselin Georgiev
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include <libexif/exif-data.h>
#include <time.h>

class ExifWrapper;

class FullExifData {
	char **tags;
	char **vals;
	int c;
public:
	FullExifData();
	~FullExifData();
	int count() const;
	const char * get_tag_name(int idx);
	const char * get_tag_value(int idx);
	friend class ExifWrapper;
};

struct Rational {
	int num, den;
	Rational() {}
	Rational(int n, int d) : num(n), den(d) {}
	void norm();
	bool operator == (const Rational & rhs) const;
	bool operator != (const Rational & rhs) const;
	bool operator < (const Rational &rhs) const;
	bool operator <= (const Rational & rhs) const;
	bool operator > (const Rational &rhs) const;
	bool operator >= (const Rational & rhs) const;
	int round() const;
	operator float () const;
};

class ExifWrapper {
	ExifData *data;
	bool get_tag(int id, void *result, int size);
public:
	ExifWrapper();
	~ExifWrapper();
	bool try_load(const char* fn);
	bool try_save(const char* fn);
	bool has_exif() const;
	// utility functions:
	
	/**
	 * get_rotation() - gets the stored EXIF info for picture rotation
	 * possible return values - [-1..3]
	 * @retval -1  - the info is not available
	 * @retval 0..3 - the rotation in the same format as RawImg::rotation field
	 */
	int get_rotation(); 

	/**
	 * get_datetime() - gets the stored EXIF Date/Time of when the image was taken
	 * @param result - [out] the result is written here, in UNIX time_t format
	 * @retval true - success
	 * @retval false - no exif, no timedate field in the exif, etc.
	 */
	bool get_datetime(time_t &result);
	
	/**
	 * get_exposure() - returns the exposure time as a rational number
	 * @param result - [out] the result is written here
	 * @retval true - success
	 * @retval false - some error, or exposure data is not present
	 * NOTE: Uses ExposureTime exif field. If not present, falls back to 
	 *       ShutterSpeed.
	 */
	bool get_exposure(Rational & result);
	
	/**
	 * get_aperture() - gets the aperture value (F-number).
	 * @param result - [out] f-number (e.g. if aperture is f/2.8, then result=2.8)
	 * @returns true on success, false on error or if data is missing
	 * NOTE: Uses exif.FNumber field. If not present, falls back to Aperture.
	 */
	bool get_aperture(float &result);
	
	/**
	 * get_iso() - gets the ISO Speed rating
	 * @param result - [out] the iso speed rating
	 * @retval true - success
	 * @retval false - failure or data is missing
	 * NOTE: Uses exif.IsoSpeedRating field.
	 */
	bool get_iso(int & result);
	
	/**
	 * get_focal_length() - gets the real focal length
	 * @param result - [out] the focal length, in mm
	 * @retval true - success
	 * @retval false - failure or data is missing
	 * NOTE: Uses exif.FocalLength field.
	 */
	bool get_focal_length(int & result);
	
	/**
	 * get_effective_focal_length() - gets the effective (35mm film eqv.) focal length
	 * @param result - [out] the effective focal length, in mm
	 * @retval true - success
	 * @retval false - failure or data is not present
	 * NOTE: Uses exif.FocalLengthIn35mmFilm
	 *
	 * In detail: most digital cameras have small sensors, compared to the 
	 * 35mm film frame area. These cameras need shorter focal lengths to 
	 * get the same effect as the 35mm film lenses. E.g. the Canon 350 D has
	 * 1.6 "Crop" factor, which means, that a lens of X mm focal length will 
	 * behave as 1.6 * X mm one. E.g. if you mount the EF 28mm lens on the
	 * Canon EOS 350D, the exif data for any picture, taken with it will show:
	 * 
	 * get_focal_length() = 28
	 * get_effective_focal_length() = 45
	 *
	 * Since there's no common crop factor (or even a way to mark it in Exif),
	 * we hope that the FocalLengthIn35mmFilm field will be present.
	 */
	bool get_effective_focal_length(int & result);
	
	/**
	 * get_full_exif() - gets an array of all Exif tags, stored in the image
	 * @param result - where the result will be written
	 * @retval true on success, false if there's no exif data, etc.
	 */
	bool get_full_exif(FullExifData & result);
};
