/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <stdio.h>
#include <vector>
#include <string>
#include <algorithm>
#include <math.h>
#include "bitmap.h"
#include "tools.h"
#include "parser.h"

using namespace std;

static vector<string> get_all_readable_images_in_dir(void)
{
	vector<string> res;
	char *exts[] = { "*.jpg", "*.bmp", "" };
	for (int i = 0; exts[i][0]; i++) {
		FindData dta(exts[i]);
		for (int j = 0; j < dta.size(); j++)
			res.push_back(dta[j]);
	}
	
	return res;
}

static int xround(float x)
{
	return (int) floor(x + 0.5f);
}

static void resize_if_needed(RawImg &img)
{
	int largest = std::max(img.get_x(), img.get_y());
	if (largest <= resize)
		return;
	
	float factor = resize / (float) largest;
	img.Resize(xround(img.get_x() * factor), xround(img.get_y() * factor));
}

Rgb operator * (const Rgb &a, const Rgb &b)
{
	Rgb res;
	res.b = a.b * b.b;
	res.g = a.g * b.g;
	res.r = a.r * b.r;
	return res;
}

Rgb operator + (const Rgb &a, const Rgb &b)
{
	Rgb res;
	res.b = a.b + b.b;
	res.g = a.g + b.g;
	res.r = a.r + b.r;
	return res;
}

static string filter_name(string s)
{
	for (unsigned i = 0; i < s.length(); i++) {
		if ((s[i] >= 'a' && s[i] <= 'z') || (s[i] >= 'A' && s[i] <= 'Z') || (s[i] >= '0' && s[i] <= '9'))
			continue;
		s[i] = '_';
	}
	return s;
}

static inline Rgb process_pixel(const Rgb& input)
{
	if (!grayscale) return input;
	float saturation = sqrtf(
		input.r * input.r * 0.241f +
		input.g * input.g * 0.691f +
		input.b * input.b * 0.068f);
	Rgb ret;
	ret.r = ret.g = ret.b = saturation;
	return ret;
}

static void make_anaglyph(RawImg & a, RawImg & b, unsigned mask_a, unsigned mask_b, string fn)
{
	RawImg c;
	c.SetSize(a.get_x(), a.get_y());
	
	Rgb mula, mulb; 
	
	mula.CreateFrom(mask_a);
	mulb.CreateFrom(mask_b);
	
	Rgb *pa = a.get_data(), *pb = b.get_data(), *pc = c.get_data();
	int size = a.get_x() * a.get_y();
	
	Task task("Creating anaglyph");
	task.set_steps(size);
	for (int i = 0; i < size; i++) {
		pc[i] = process_pixel(pa[i]) * mula + process_pixel(pb[i]) * mulb;
		if (i & 2047 == 0)
			task.set(i);
	}
	c.SaveToJPG(fn.c_str(), jpegQuality);
}

static void make_cross(RawImg & a, RawImg & b, string fn)
{
	RawImg c;
	c.SetSize(a.get_x() * 2, a.get_y());
	
	Rgb *pa = a.get_data(), *pb = b.get_data(), *pc = c.get_data();
	int x = a.get_x(), y = a.get_y();
	Task task("Creating cross image");
	task.set_steps(y);
	for (int j = 0; j < y; j++) {
		for (int i = 0; i < x; i++) {
			pc[i + x + j * 2 * x] = pa[i + j * x];
			pc[i     + j * 2 * x] = pb[i + j * x];
		}
		task++;
	}
	c.SaveToJPG(fn.c_str(), jpegQuality);
}

static void process(string fn1, string fn2)
{
	printf("Processing pair `%s' and `%s'...\n", fn1.c_str(), fn2.c_str());
	RawImg a, b;
	a.LoadFromImageFile(fn1.c_str());
	resize_if_needed(a);
	b.LoadFromImageFile(fn2.c_str());
	resize_if_needed(b);
	//
	if (a.get_x() != b.get_x() || a.get_y() != b.get_y()) {
		printf("Files `%s' and `%s' differ in dimensions, skipping!\n", fn1.c_str(), fn2.c_str());
		return;
	}
	//
	make_anaglyph(a, b, 0xff0000, 0x00ffff, "asg_ana_rc_" + filter_name(fn1) + " " + filter_name(fn2));
	make_anaglyph(a, b, 0xff0000, 0x0000ff, "asg_ana_rb_" + filter_name(fn1) + " " + filter_name(fn2));
	make_cross(a, b, "asg_cross_" + filter_name(fn1) + " " + filter_name(fn2));
	//
	progress.cleanup();
}

int main(int argc, char **argv)
{
	parser_init();
	parse_config(get_home_dir() + ".config");
	parse_config(get_home_dir() + ".asgconfig");
	parse_config(".config");
	parse_config(".asgconfig");
	if (!parse_cmdline(argc, argv)) return -1;
	//
	vector<string> names = get_all_readable_images_in_dir();
	if (names.empty()) {
		printf("Warning: no image (.jpg or .bmp) files in current dir, bailing out.\n");
		return 0;
	}
	for (unsigned i = 0; i < names.size() - 1; i += crossover ? 1 : 2) {
		string fn1 = swaplr ? names[i + 1] : names[i];
		string fn2 = swaplr ? names[i] : names[i + 1];
		process(fn1, fn2);
	}
	
	return 0;
}
