/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <stdio.h>
#include <vector>
#include <string>
#include <algorithm>
#include <math.h>
#include "bitmap.h"
#include "tools.h"
#include "parser.h"

using namespace std;

int conf_resize = 180;


static void usage(void)
{
	printf("thumbnailize - create thumbnails of all .jpg/.bmp files from the current dir\n\n");
	printf("Usage: thumbnailize [options]\n\n");
	printf("Options are:\n");
	printf("\t-h - print this help\n");
	printf("\t-resize=xxx  - resulting thumbnails will have the specified\n");
	printf("\t               length (pixels) on the larger side\n");
}

static inline char upcase(char c)
{
	if (c >= 'a' && c <= 'z') return c - 32;
	else return c;
}

static vector<string> get_all_readable_images_in_dir(void)
{
	vector<string> res;
	char *exts[] = { "*.jpg", "*.bmp", "" };
	for (int i = 0; exts[i][0]; i++) {
		FindData dta(exts[i]);
		for (int j = 0; j < dta.size(); j++)
			res.push_back(dta[j]);
	}
	
	return res;
}

static int xround(float x)
{
	return (int) floor(x + 0.5f);
}

static void resize_if_needed(RawImg &img)
{
	int largest = std::max(img.get_x(), img.get_y());
	if (largest <= conf_resize)
		return;
	
	float factor = conf_resize / (float) largest;
	img.Resize(xround(img.get_x() * factor), xround(img.get_y() * factor));
}

static bool parse_cmds(int argc, char **argv)
{
	bool setsize = false;
	for (int i = 1; i < argc; i++) {
		bool recognized = false;
		if (!strncmp(argv[i], "-resize=", 8)) {
			if (1 != sscanf(argv[i], "-resize=%d", &conf_resize) ||	conf_resize < 1 || conf_resize > 9999) {
				printf("Invalid argument: `%s'\n", argv[i]);
				return false;
			}
			setsize = true;
			recognized = true;
		}
		if (!strcmp(argv[i], "--help") || !strcmp(argv[i], "-h")) {
			usage();
			return false;
		}
		
		if (!recognized) {
			printf("Parameter unrecognized: `%s'\n", argv[i]);
			return false;
		}
	}
	if (!setsize) {
		printf("Warning: using default output resolution of %d!\n", conf_resize);
	}
	return true;
}

static string get_resulting_filename(string s)
{
	string ext = s.substr(s.length()-3);
	if (upcase(ext[0]) == 'B' && upcase(ext[1]) == 'M' && upcase(ext[2]) == 'P')
		s = s.substr(s.length()-3) + "jpg";
	return "tn_" + s;
}


int main(int argc, char **argv)
{
	if (!parse_cmds(argc, argv)) return -1;
	//
	vector<string> names = get_all_readable_images_in_dir();
	if (names.empty()) {
		printf("Warning: no image (.jpg or .bmp) files in current dir, bailing out.\n");
		return 0;
	}
	for (unsigned i = 0; i < names.size() ; i++) {
		RawImg a;
		if (names[i].substr(0, 3) == "tn_") {
			printf("Skipping `%s' - probably a thumbnail already\n", names[i].c_str());
			continue;
		}
		printf("Resizing %s -> tn_%s\n", names[i].c_str(), names[i].c_str());
		a.LoadFromImageFile(names[i].c_str());
		resize_if_needed(a);
		string outfn = get_resulting_filename(names[i]);
		a.SaveToJPG(outfn.c_str(), 85);
		progress.cleanup();
	}
	
	return 0;
}
