/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include "tools.h"

#ifdef _WIN32
#include <windows.h>
#include <vector>
#include <stdlib.h>
#else
#include <glob.h>
#include <sys/time.h>
#endif

using namespace std;


/**
 * @class FindData
 */
void FindData::_construct()
{
	_names = NULL;
	_count = 0;
	_iterator = 0;
}

#ifdef _WIN32
static int cmp_charp(const void *a, const void *b)
{
	const char *x = *(const char **) a, *y = *(const char **) b;
	return strcmp(x, y);
}
#endif

bool FindData::_populate(const char* path)
{
#ifdef _WIN32
//	win32's find does not return the full path name, so extract the root first
	int rootLen = strlen(path)-1;
	while (rootLen >= 0 && path[rootLen] != '/' && path[rootLen] != '\\') rootLen--;
	rootLen++;
//

	vector<char*> all;
	WIN32_FIND_DATA findData;
	HANDLE fd = FindFirstFile(path, &findData);
	if (fd == INVALID_HANDLE_VALUE) return false;
	do {
		if (findData.cFileName[0] == '.' &&
		(findData.cFileName[1] == 0 || findData.cFileName[1] == '.')) continue;
		int fileLen=strlen(findData.cFileName);
		int outlen = 1 + fileLen + rootLen;
		char *s = new char[outlen];
		if (!s) continue;
		memcpy(s, path, rootLen);
		memcpy(s+rootLen, findData.cFileName, fileLen+1);
		all.push_back(s);
	} while ( 0 != FindNextFile(fd, &findData));
	FindClose(fd);
	qsort(&all[0], all.size(), sizeof(char*), cmp_charp);
	//
	_names = new char*[all.size()];
	for (unsigned i = 0; i < all.size(); i++)
		_names[i] = all[i];
	_count = (int) all.size();
#else
	glob_t pg;
	if (0 != glob(path, GLOB_MARK|GLOB_TILDE, NULL, &pg)) return false;
	_names = new char*[pg.gl_pathc];
	_count = pg.gl_pathc;
	for (int i = 0; i < _count; i++) {
		_names[i] = new char[1 + strlen(pg.gl_pathv[i])];
		strcpy(_names[i], pg.gl_pathv[i]);
	}
	globfree(&pg);
#endif
	_iterator = 0;
	return true;
}

void FindData::_destruct(void)
{
	if (_count) {
		for (int i = 0; i < _count; i++)
			delete [] _names[i];
		delete [] _names;
		
	}
	_names = NULL; _count = 0;
}

FindData::FindData()
{
	_construct();
}

FindData::~FindData()
{
	_destruct();
}

FindData::FindData(const char *path)
{
	_construct();
	_populate(path);
}

bool FindData::search(const char *path)
{
	return _populate(path);
}

bool FindData::empty(void) const
{
	return _count == 0;
}

int FindData::size(void) const
{
	return _count;
}

const char* FindData::operator [](int index) const
{
	if (index < 0 || index >= _count || _names == NULL) return NULL;
	return _names[index];
}

void FindData::reset(void)
{
	_iterator = 0;
}

const char* FindData::next(void) 
{
	if (_iterator >= _count || _names == NULL) return NULL;
	return _names[_iterator++];
}

/**
 * @Class Progress
*/ 

Progress::Progress()
{
	last_time = -1;
}

void Progress::task_set_name(const char * name)
{
	strcpy(taskid, name);
	prog = 0;
	cleanup(true);
	update(true);
}

void Progress::task_set_progress(double x)
{
	prog = x;
	update();
}

void Progress::task_finish(void)
{
	prog = 1;
	update(true);
}

void Progress::cleanup(bool flush)
{
	clear_line();
	if (flush) fflush(stdout);
}

void Progress::update(bool now)
{
	if (!now) {
		if (xtime() - last_time < 0.2) return;
	}
	last_time = xtime();
	double x = prog;
	if (x > 1.0) x = 1.0;
	if (x < 0.0) x = 0.0;
	printf("\r[%s ... %6.2lf%%]", taskid, prog * 100.0);
	fflush(stdout);
}

Progress progress;

/**
 * @class Task
*/
Task::Task(const char * name)
{
	progress.task_set_name(name);
	steps = 1;
	cstep = 0;
}

Task::~Task()
{
	progress.task_finish();
}

void Task::set_steps(int steps)
{
	this->steps = steps;
}

void Task::operator ++ (int)
{
	cstep++;
	progress.task_set_progress(cstep / (double) steps);
}

void Task::set(int x)
{
	cstep = x;
	progress.task_set_progress(cstep / (double) steps);
}

/*---------------------------------------------------------------------*/

static string capitalize(string x)
{
	for (unsigned i = 0; i < x.length(); i++)
		if (x[i] >= 'a' && x[i] <= 'z') x[i] -= 32;
	return x;
}

bool ExtensionIs(string file, string ext)
{
	ext = capitalize(ext);
	file = capitalize(file);
	unsigned L1 = file.length(), L2 = ext.length();
	for (unsigned i = 1; i <= L2; i++)
		if (file[L1-i] != ext[L2-i]) return false;
	return true;
}

string get_home_dir(void)
{
#ifdef _WIN32
#warning this could be handled more gracefully
	return "C:\\";
#else
	const char * s = getenv("HOME");
	if (!s) return "";
	else return string(s) + "/";
#endif
}

Uint32 ByteSwap32(Uint32 x)
{
	x = ((x & 0xffff) << 16) | ((x >> 16) & 0xffff);
	x = ((x & 0x00ff00ff) << 8) | ((x >> 8) & 0x00ff00ff);
	return x;
}
Uint16 ByteSwap16(Uint16 x)
{
	x = ((x & 0xff) << 8) | ((x >> 8) & 0xff);
	return x;
}

void clear_line(void)
{
	printf("\r");
	for (int i = 0; i < 8; i++)
		printf("        ");
	printf("\r");
}

double xtime(void)
{
#ifdef _WIN32
	unsigned x = GetTickCount();
	static unsigned startx = 0;
	if (!startx) {
		startx = x; return 0;
	} else {
		return (x - startx) * 0.001;
	}
#else
	struct timeval t;
	gettimeofday(&t, NULL);
	static timeval start = {0, 0};
	if (!start.tv_sec) {
		start = t;
		return 0;
	} else {
		return (t.tv_sec - start.tv_sec) + (t.tv_usec - start.tv_usec) * 0.000001;
	}
#endif
}
