#ifndef __TOOLS_H__
#define __TOOLS_H__

#include <string>

typedef unsigned short int Uint16;
typedef unsigned int Uint32;

/**
 * @class  FindData
 * @brief  Generic class for searching directories for files
 * @date   2006-06-12
 * @author Veselin Georgiev
 *
 * The class is built upon FindFirstFile()/FindNextFile() under Win32
 * and upon glob() on *nix. It creates a list of files, matching
 * the given pattern.
 *
 * When searching in a distant path (e.g. "../sources/ *.so"), the returned
 * results are prefixed with that path (e.g. "../sources/foo.so", "../sources/bar.so", etc)
*/ 
class FindData {
	char **_names;
	int _count;
	int _iterator;
	
	void _construct();
	void _destruct();
	bool _populate(const char *path);
public:
	FindData();
	~FindData();
	
	/// Match the given path and wildcard for files (actually, it calls search())
	FindData(const char *path);
	
	/// Match the given path and wildcard for files
	/// @return true on success, false on failure
	bool search(const char *path);
	
	/// True if no files were found
	bool empty(void) const;
	
	/// The number of the files found
	int size(void) const;
	
	/// Returns the next matching file; or NULL if there are no more
	const char *next(void);
	
	/// Resets the current file iterator to the beginning
	void reset(void);
	
	/// Return a matching file by index [0 .. size() - 1]
	const char *operator[] (int index) const;
};

class Progress {
	char taskid[80];
	double prog;
	double last_time;
	void update(bool forced = false);
public:
	Progress();
	void task_set_name(const char *);
	void task_set_progress(double prog);
	void task_finish(void);
	void cleanup(bool flush = false);
};

extern Progress progress;

class Task {
	int steps, cstep;
public:
	Task(const char * name);
	~Task();
	void set_steps(int steps);
	void operator ++ (int);
	void set(int cstep);
};

bool ExtensionIs(std::string file, std::string ext);

std::string get_home_dir(void);

Uint32 ByteSwap32(Uint32);
Uint16 ByteSwap16(Uint16);

void clear_line(void);

double xtime(void);

#if __BYTE_ORDER == __LITTLE_ENDIAN
#	define itohl(x) (x)
#	define itohs(x) (x)
#	define htoil(x) (x)
#	define htois(x) (x)
#else
#	define itohl(x) ByteSwap32(x)
#	define itohs(x) ByteSwap16(x)
#	define htoil(x) ByteSwap32(x)
#	define htois(x) ByteSwap16(x)
#endif


#endif // __TOOLS_H__
