/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <sstream>
#include <string>
#include <vector>
#include "exif.h"
#include "jpeg-data.h"

/* -- we need this struct for internal workings */
struct ExifTempStruct {
	int id;
	void *result;
	int size;
	bool found;

	ExifTempStruct(int ii, void *res, int sz): id(ii), result(res), size(sz) 
	{ found = false; }
};

/* We need rational numbers for Exif structures - some of them are rationals */
static inline int gcd(int a, int b)
{
	while (a != 0 && b != 0) {
		a %= b;
		if (a) b %= a;
	}
	return a + b;
}

void Rational::norm()
{
	int g = gcd(num, den);
	if (g > 1) {
		num /= g;
		den /= g;
	}
}

bool Rational::operator == (const Rational & rhs) const
{
	return (long long) num * rhs.den == (long long) den * rhs.num;
}
bool Rational::operator != (const Rational & rhs) const
{
	return (long long) num * rhs.den != (long long) den * rhs.num;
}
bool Rational::operator <  (const Rational & rhs) const
{
	return (long long) num * rhs.den < (long long) den * rhs.num;
}
bool Rational::operator <= (const Rational & rhs) const
{
	return (long long) num * rhs.den <= (long long) den * rhs.num;
}
bool Rational::operator >  (const Rational & rhs) const
{
	return (long long) num * rhs.den > (long long) den * rhs.num;
}
bool Rational::operator >= (const Rational & rhs) const
{
	return (long long) num * rhs.den >= (long long) den * rhs.num;
}

Rational::operator float() const
{
	if (den == 0) return 0;
	return (float) num / (float) den;
}

int Rational::round() const
{
	if (den == 0) return 0;
	return (num + den / 2) / den;
}


/* The actual ExifWrapper class follows */
ExifWrapper::ExifWrapper()
{
	data = NULL;
}

ExifWrapper::~ExifWrapper()
{
	if (data)
		exif_data_free(data);
	data = NULL;
}

bool ExifWrapper::try_load(const char *fn)
{
	data = exif_data_new_from_file(fn);
	return (data != NULL);
}

bool ExifWrapper::try_save(const char *fn)
{
	if (!data) return false;
	unsigned char *binary_data = NULL;
	unsigned int binary_len = 0;
	exif_data_save_data(data, &binary_data, &binary_len);
	if (!binary_data || !binary_len) {
		return false;
	}
	JPEGData *jdata;

	/* Parse the JPEG file */
	jdata = jpeg_data_new_from_file (fn);
	if (!jdata) {
		return false;
	}

	jpeg_data_set_exif_data (jdata, data);

	/* Save the modified image. */
	jpeg_data_save_file (jdata, fn);
	jpeg_data_unref (jdata);

	return true;
}

static void entry_fun(ExifEntry *entry, void *data)
{
	ExifTempStruct *t = static_cast<ExifTempStruct*>(data);
	if (t->found) return;
	if (t->id == entry->tag) {
		if (entry->data) {
			t->found = true;
			memcpy(t->result, entry->data, t->size);
		}
	}
}

static void content_fun(ExifContent* content, void *data)
{
	exif_content_foreach_entry(content, entry_fun, data);
}

bool ExifWrapper::get_tag(int id, void *result, int size)
{
	ExifTempStruct st(id, result, size);
	exif_data_foreach_content(data, content_fun, &st);
	return st.found;
}

int ExifWrapper::get_rotation()
{
	if (!data) return -1;
	int result = 0;
	if (!get_tag(0x0112, &result, 1)) return -1;
	if (!result) return -1;
	
	switch (result) {
		case 1: return 0;
		case 3: return 2;
		case 6: return 3;
		case 8: return 1;
	}
	
	return -1;
}

bool ExifWrapper::get_datetime(time_t &result)
{
	if (!data) return false;
	char datestring[24];
	if (!get_tag(0x0132, datestring, 20)) return false;
	// datefmt = "YYYY:MM:DD HH:MM:SS"
	int integers[6] = {1900, 1, 1, 0, 0, 0};
	int num_ints = 0;
	for (int i = 0; i < (int) strlen(datestring); i++)
		if (datestring[i] == ':')
			datestring[i] = ' ';
	std::stringstream ss(datestring);
	int x;
	while ((ss >> x) && (num_ints < 6)) {
		integers[num_ints++] = x;
	}
	if (num_ints == 0) return false;
	int year, month, day, hour, minute, second;
	year = integers[0];
	if (year < 1900 || year > 5000) return false;
	month = integers[1];
	if (month < 1 || month > 12) return false;
	day = integers[2];
	if (day < 1 || day > 31) return false;
	hour = integers[3];
	if (hour < 0 || hour > 23) return false;
	minute = integers[4];
	if (minute < 0 || minute > 59) return false;
	second = integers[5];
	if (second < 0 || second > 60 /* allow leap seconds */) return false;
	//
	struct tm t;
	t.tm_year = year - 1900;
	t.tm_mon = month - 1;
	t.tm_mday = day;
	t.tm_hour = hour;
	t.tm_min = minute;
	t.tm_sec = second;
	result = mktime(&t);
	return true;
}

bool ExifWrapper::get_exposure(Rational & result)
{
	if (!data) return false;
	if (!get_tag(0x829A, &result, sizeof(result))) {
		// try the other one
		if (!get_tag(0x9201, &result, sizeof(result)))
			return false;
	}
	return true;
}

bool ExifWrapper::get_aperture(float &result)
{
	if (!data) return false;
	Rational r;
	if (get_tag(0x829D, &r, sizeof(r))) {
		result = r;
		return true;
	}
	if (get_tag(0x9202, &r, sizeof(r))) {
		result = r;
		return true;
	}
	return false;
}

bool ExifWrapper::get_iso(int & result)
{
	if (!data) return false;
	short a;
	if (!get_tag(0x8827, &a, sizeof(a)))
		return false;
	result = (int) a;
	return true;
}

bool ExifWrapper::get_focal_length(int & result)
{
	if (!data) return false;
	Rational r;
	if (!get_tag(0x920A, &r, sizeof(r)))
		return false;
	result = r.round();
	return true;
}

bool ExifWrapper::get_effective_focal_length(int & result)
{
	if (!data) return false;
	short s;
	if (!get_tag(0xa405, &s, sizeof(s)))
		return false;
	result = (int) s;
	return true;
}


struct GetFullExifClosure {
	std::vector<std::pair<std::string, std::string> > data;
};

static void full_exif_entry_fun(ExifEntry *entry, void *data)
{
	GetFullExifClosure *t = static_cast<GetFullExifClosure*>(data);
	char temp1[1024], temp2[1024];
	strncpy(temp1, exif_tag_get_title(entry->tag), sizeof(temp1));
	exif_entry_get_value(entry, temp2, 1024);
	t->data.push_back(std::make_pair(std::string(temp1), std::string(temp2)));
}

static void full_exif_content_fun(ExifContent* content, void *data)
{
	exif_content_foreach_entry(content, full_exif_entry_fun, data);
}


bool ExifWrapper:: get_full_exif(FullExifData & result)
{
	if (!data) return false;
	GetFullExifClosure temp;
	exif_data_foreach_content(data, full_exif_content_fun, &temp);
	result.c = (int) temp.data.size();
	if (result.c) {
		int n = result.c;
		result.tags = (char **)  malloc(sizeof(char*) * n);
		result.vals = (char**) malloc(sizeof(char*) * n);
		for (int i = 0; i < n; i++) {
			result.tags[i] = (char*) malloc(temp.data[i].first.length()+1);
			strcpy(result.tags[i], temp.data[i].first.c_str());
			result.vals[i] = (char*) malloc(temp.data[i].second.length()+1);
			strcpy(result.vals[i], temp.data[i].second.c_str());
		}
	}
	return true;
}

/**
 * @class FullExifData
 */

FullExifData::FullExifData()
{
	tags = vals = NULL;
	c = 0;
}

FullExifData::~FullExifData()
{
	if (tags) {
		for (int i = 0; i < c; i++)
			if (tags[i]) {
				free(tags[i]);
				tags[i] = NULL;
			}
	}
	free(tags);
	//
	if (vals) {
		for (int i = 0; i < c; i++) {
			if (vals[i]) {
				free(vals[i]);
				vals[i] = NULL;
			}
		}
	}
	free(vals);
}

int FullExifData::count() const
{
	return c;
}


const char * FullExifData::get_tag_name(int idx)
{
	return (idx >= 0 && idx < c) ? tags[idx] : NULL;
}

const char * FullExifData::get_tag_value(int idx)
{
	return (idx >= 0 && idx < c) ? vals[idx] : NULL;
}
