/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <wx/wx.h>

#include "main.h"

volatile int spinlk;
const double MinimumTimeToShowETA=2.0; // measured in seconds
const double TimerUpdateInterval=0.1;  // measured in seconds

class ProgressDialog : public wxDialog {
public:
	ProgressDialog();
	void SetProgress(double);
	void SetProgressText(const wxString &, int hint);
	double prog, last_prog;
	int last_hint;
	wxString txt;
	bool ended;
	int endres;

	void OnProgressTimer(wxTimerEvent& event);
	void StopTimer();
private:

	wxString MakeETA(double time_elapsed);

	wxGauge *Gauge;
	wxStaticText *TheText;
	wxMutex DrawCS;
	wxTimer * ProgTimer;
	wxStaticText *Percentile, *ETA;
	int ticks;

	DECLARE_EVENT_TABLE()
};


BEGIN_EVENT_TABLE(ProgressDialog, wxDialog)
	EVT_TIMER(1, ProgressDialog:: OnProgressTimer)
END_EVENT_TABLE()

#if defined _MSC_VER || defined __APPLE__
#define PROG_YSIZE 160
#else
#define PROG_YSIZE 110
#endif


ProgressDialog:: ProgressDialog() 
		: wxDialog(globMainFrame, 
		wxID_ANY, 
		"Please wait, working...",
		wxDefaultPosition,
		wxSize(400, PROG_YSIZE)
		)
{
	prog = last_prog = 0.0;
	last_hint = -1;
	/* Set up the GUI */
	TheText = new wxStaticText(this, -1, wxT("Initializing..."),
		wxPoint(10, 24));
	Gauge = new wxGauge(this, -1, 1000, wxPoint(10, 50),
		wxSize(380, 20), wxGA_HORIZONTAL | wxGA_SMOOTH);
	Percentile = new wxStaticText(this, -1, wxT("0 %"),
		wxPoint(12, 80));
	wxFont txtFont = Percentile->GetFont();
	txtFont.SetWeight(wxFONTWEIGHT_BOLD);
	Percentile->SetFont(txtFont);
	ETA = new wxStaticText(this, -1, wxT("Estimated time remaining: unknown"),
		wxPoint(100, 80));
	ETA->Hide();

	/* Set up the timer */
	ticks = 0; last_prog = -1; prog = 0;
	ended = false; endres = -1;
	ProgTimer = new wxTimer(this, 1);
	ProgTimer->Start((int) (1000 * TimerUpdateInterval));
	
}

void ProgressDialog:: StopTimer()
{
	ProgTimer->Stop();
	delete ProgTimer;
}

void ProgressDialog:: SetProgress(double new_prog)
{
	DrawCS.Lock();
	if (new_prog >= 0 && new_prog <= 1.0 && new_prog > prog) {
		prog = new_prog;
	}
	DrawCS.Unlock();
}

void ProgressDialog:: SetProgressText(const wxString & thetext, int hint)
{
	DrawCS.Lock();
	if (hint > last_hint) {
		txt = thetext;
		last_hint = hint;
	}
	DrawCS.Unlock();
}

void ProgressDialog:: OnProgressTimer(wxTimerEvent& event)
{
	if (ended) {
		StopTimer();
		spinlk = 1;
		EndModal(endres);
		spinlk = 0;
		return;
	}
	ticks++;
	if (prog == last_prog) return;
	last_prog = prog;
	
	DrawCS.Lock();
	Gauge->SetValue((int) (1000.0 * prog));	
	TheText->SetLabel(txt);
	Percentile->SetLabel(wxString::Format(wxT("%d %%"), (int) (100.0 * prog)));
	double time_elapsed = ticks * TimerUpdateInterval;
	if (time_elapsed > MinimumTimeToShowETA && prog > 0.0) {
		ETA->SetLabel(wxT("Estimated time remaining: ") + 
			MakeETA(time_elapsed));
		if (!ETA->IsShown())
			ETA->Show();
	}
	DrawCS.Unlock();
	
}

wxString ProgressDialog:: MakeETA(double time_elapsed)
{
	double time_rem = (1-prog) / prog * time_elapsed;
	if (time_rem > 86400)
		time_rem = 86400; // limit to one day
	int seconds = (int) time_rem;
	if (seconds < 60)
		return wxString::Format(wxT("%d seconds."), seconds);
	if (seconds < 3600)
		return wxString::Format(wxT("%dm %ds."), seconds/60, seconds % 60);
	return wxString::Format(wxT("%d:%02d:%02d."), seconds / 3600, seconds % 3600
		/ 60, seconds % 60);
}

ProgressDialog *PDlg;

void progress_start(void)
{
	PDlg = new ProgressDialog;
}

void progress_stop(void)
{
	if (PDlg) {
		PDlg->Destroy();
		PDlg = NULL;
	}
}

int progress_enter(void)
{
	int res = PDlg->ShowModal();
	while (spinlk) wxMilliSleep(333);

	PDlg->Destroy();
	PDlg = NULL;
	return res;
}

void progress_end(int result)
{
	if (PDlg)
	{
		PDlg->endres = result;
		PDlg->ended = true;
	}
}

void progress_set_progress(double prog)
{
	if (PDlg) {
		PDlg->SetProgress(prog);
	}
}

void progress_set_text(const wxString & thetext, int hint)
{
	if (PDlg) {
		PDlg->SetProgressText(thetext, hint);
	}
}
