/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <wx/wx.h>
#include <wx/image.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <vector>
#include <algorithm>
using namespace std;

#include "project.h"
#include "progress.h"
#include "preferences.h"
#include "tools.h"
#include "defaults.h"
#include "bmphandlingdialog.h"

const char *BmpHandlingStr[] = {
	"Undefined",
	"Ignore",
	"ToJPG",
	"ToBMP",
	""
};

extern wxString FontChoices[];

// Load the default values for a ProjectInfo Structure

void ProjectInfo :: Defaults()
{
	Title = DEFAULT_TITLE;
	iPath = DEFAULT_INPUT_PATH;
	oPath = DEFAULT_OUTPUT_PATH;
	PicSize = DEFAULT_PIC_SIZE;
	ThumbSize = DEFAULT_THUMB_SIZE;
	has_frame = DEFAULT_HAS_FRAME;
	ThumbFormat = FORMAT_JPEG;
	BackgroundColor = DEFAULT_BACKGROUND_COLOR;
	TextColor = DEFAULT_TEXT_COLOR;
	PicsPerRow = DEFAULT_IMAGES_PER_ROW;
	PicsPerGallery = DEFAULT_IMAGES_PER_GALLERY;
	InfoPerImg = DEFAULT_INFO_PER_IMG;
	JpegQuality = DEFAULT_JPEG_QUALITY;
	BackgroundImage = DEFAULT_BACKGROUND_IMAGE;
	Rollover = DEFAULT_ROLLOVER_STATE;
	BmpHandling = BMP_HANDLING_UNDEFINED;
	galleryType = GT_PAGE_WITH_THUMBS;
	font = wxT(DEFAULT_FONT_CHOICE);
	exif_stats = DEFAULT_EXIF_STATS;
}


/*****************************************************************************/
/** @class Project                                                          **/
/*****************************************************************************/

Project:: Project()
{
	count = 0;
	thumbs = NULL;
	state = NULL;
	FilenameToSave = wxEmptyString;
	pi.Defaults();
}

Project:: ~Project ()
{
	if (thumbs)
		Destroy();
	if (state)
		Destroy();
}

void Project:: Destroy()
{
	if (thumbs) delete [] thumbs;
	if (state) delete [] state;
	thumbs = NULL;
	state = NULL;
	count = 0;
}

void Project:: FreeImageData()
{
	for (int i = 0; i < count; i++) 
		thumbs[i].Destroy();
}

bool Project:: RecreateThumbs(void)
{
	return RunThread(TRD_RECREATE_THUMBS);
}

bool Project:: CreateGallery(bool regenerate_pics)
{
	if (regenerate_pics) {
		return RunThread(TRD_GENERATE_GALLERY_TRUE);
	} else {
		return RunThread(TRD_GENERATE_GALLERY_FALSE);
	}
}

bool Project:: RunThread(int what)
{
	counter.reset(); doneCounter.reset();
	progress_start();
	
	int threadcount = getCPUCount();
	if (threadcount < 1 || !preferences.smp) threadcount = 1;
		
	SharedCounter thread_counter;
	thread_counter.set(threadcount);
		
	if (what == TRD_GENERATE_GALLERY_TRUE || what == TRD_GENERATE_GALLERY_FALSE) {
		if (!CreateGallery_init()) return false;
	}
	
	for (int i = 0; i < threadcount; i++) {
		int code = what;
		if (i == 0) code |= 0x80000000;
		DoWorkThread * thread = new DoWorkThread(this, code, &thread_counter);
		thread->Create();
		thread->Run();
	}
	
	int res = progress_enter();
	return (res==1);
}

static bool exif_cmp(const RawImg &a, const RawImg &b)
{
	return a.ExifDateTime() < b.ExifDateTime();
}

bool Project:: NewProject(const ProjectInfo & ProjInfo)
{
	if (thumbs) Destroy();
	count = 0;
	pi = ProjInfo;
	vector<wxString> FileNames;
	const char *ExtensionsSupported[] = { "*.jpg", "*.bmp", "" };
	
	for (int ext = 0; ExtensionsSupported[ext][0]; ext++) {
		wxString ipp = SlashExtend(pi.iPath) + ExtensionsSupported[ext];
		wxString f = wxFindFirstFile(ipp.c_str(), wxFILE);
		int thiscount = 0;
		while (!f.empty()) {
			count++;
			thiscount++;
			FileNames.push_back(f);
			f = wxFindNextFile();
		}
		// under UNIX, handle those ugly uppercase names
#if defined __linux__ || defined unix	
		ipp = SlashExtend(pi.iPath) + Capitalize(ExtensionsSupported[ext]);
		f = wxFindFirstFile(ipp.c_str(), wxFILE);
		while (!f.empty()) {
			count++;
			thiscount++;
			FileNames.push_back(f);
			f = wxFindNextFile();
		}
#endif
		if (thiscount != 0 && !strcmp("*.bmp", ExtensionsSupported[ext])) {
			BmpHandlingDialog *bhd = new BmpHandlingDialog(&pi.BmpHandling);
			bhd->ShowModal();
			bhd->Destroy();
			if (pi.BmpHandling <= BMP_HANDLING_IGNORE) {
				FileNames.erase(FileNames.end() - thiscount,
					FileNames.end());
				count -= thiscount;
			}
		}
	}
	if (!count) {
		wxMessageBox(wxT("No Image files in this directory"), 
			wxT("Error"), wxICON_ERROR);
		return false;
	}
	sort(FileNames.begin(), FileNames.end());
	thumbs = new RawImg [count];
	for (int i =0 ; i< count; i++) {
		thumbs[i].FileName = FileNames[i];
		thumbs[i].outFileName = GetFn(FileNames[i]);
	}
	if (preferences.sort_exif) {
		int count_exifs = 0;
		for (int i = 0; i < count; i++) {
			bool res = thumbs[i].ExtractExifDateTime();
			if (res) 
				count_exifs++;
		}
		if (count_exifs > 1) {
			stable_sort(thumbs, thumbs + count, exif_cmp);
		}
	}
	state = new bool[count];
	for (int i = 0; i < count; i++)
		state[i] = true;
	MustSave = true;
	return true;
}

int Project:: size() const
{
	return count;
}

int Project:: GetCount() const
{
	int c=0, i;
	for (i = 0; i < count; i++)
		if (state[i]) c++;
	return c;
}

void Project:: SetCurrent(int new_current)
{
	current = new_current;
	if (current > count - 1)
		current = 0;
	if (current < 0)
		current = count - 1;
}

void Project:: AdjustCurrent(int adjustment)
{
	SetCurrent(current + adjustment);
}

RawImg * Project:: GetCurrentImg() const
{
	return thumbs + current;
}

bool Project:: GetCurrentState() const
{
	return state[current];
}

void Project:: SetCurrentState(bool Include)
{
	state[current] = Include;
}

int Project:: GetCurrentID() const
{
	return current;
}

wxString Project:: Get_iPath() const { return pi.iPath; }
bool Project:: GetHasFrame() const { return pi.has_frame; }
void Project:: SetHasFrame(bool new_val)
{
	pi.has_frame = new_val;
}
void Project:: SetProjectInfo(const ProjectInfo &new_pi)
{
	pi = new_pi;
}

ProjectInfo Project:: GetProjectInfo() const { return pi; }

bool Project:: LoadFromFile(const wxString & fn, bool recreate_thumbs, bool go)
{
	FILE *f;
	f = fopen(fn.c_str(), "rt");
	if (!f) return false;

	count = -1;
	pi.BackgroundImage = "";

	char line[512], field[512], value[512];
	while (fgets(line, 512, f)) {
		if (parse_pair(line, field, value)) {
			if (!strcmp(field, "Title")) {
				pi.Title = value;
			}
			if (!strcmp(field, "InputPath")) {
				pi.iPath = value;
			}
			if (!strcmp(field, "OutputPath")) {
				pi.oPath = value;
			}
			if (!strcmp(field, "PicSize")) {
				int r = sscanf(value, "%d", &pi.PicSize);
				if (1 != r) {
					fclose(f);
					return false;
				}
			}
			if (!strcmp(field, "ThumbSize")) {
				int r = sscanf(value, "%d", &pi.ThumbSize);
				if (1 != r) {
					fclose(f);
					return false;
				}
			}
			if (!strcmp(field, "HasFrame")) {
				int b;
				int r = sscanf(value, "%d", &b);
				if (1 != r) {
					fclose(f);
					return false;
				}
				pi.has_frame = (b!=0);
			}
			if (!strcmp(field, "Rollover")) {
				int b;
				int r = sscanf(value, "%d", &b);
				if (1 != r) {
					fclose(f);
					return false;
				}
				pi.Rollover = (RolloverState)b;
			}
			if (!strcmp(field, "BmpHandling")) {
				bool found = false;
				for (int i = 0; BmpHandlingStr[i][0]; i++)
					if (!strcmp(value, BmpHandlingStr[i])) {
						found = true;
						pi.BmpHandling = (BmpHandlingType)i;
						break;
					}
				if (!found) {
					fclose(f);
					return false;
				}
			}
			if (!strcmp(field, "ThumbFormat")) {
				if (!strcmp(value, "PNG"))
					pi.ThumbFormat = FORMAT_PNG;
				else
					pi.ThumbFormat = FORMAT_JPEG;
			}
			if (!strcmp(field, "BackgroundColor")) {
				int r = sscanf(value, "#%x", &pi.BackgroundColor);
				if (1 != r) {
					fclose(f);
					return false;
				}
			}
			if (!strcmp(field, "TextColor")) {
				int r = sscanf(value, "#%x", &pi.TextColor);
				if (1 != r) {
					fclose(f);
					return false;
				}
			}
			if (!strcmp(field, "PicsPerRow")) {
				int r = sscanf(value, "%d", &pi.PicsPerRow);
				if (1 != r) {
					fclose(f);
					return false;
				}
			}
			if (!strcmp(field, "PicsPerGallery")) {
				int r = sscanf(value, "%d", &pi.PicsPerGallery);
				if (1 != r) {
					fclose(f);
					return false;
				}
			}
			if (!strcmp(field, "InfoPerImg")) {
				int r = sscanf(value, "%d", &pi.InfoPerImg);
				if (1 != r) {
					fclose(f);
					return false;
				}
			}
			if (!strcmp(field, "JpegQuality")) {
				int r = sscanf(value, "%d", &pi.JpegQuality);
				if (1 != r) {
					fclose(f);
					return false;
				}
			}
			if (!strcmp(field, "BackgroundImage")) {
				pi.BackgroundImage = value;
			}
			if (!strcmp(field, "GalleryType")) {
				if (!strcmp(value, "PageWithThumbs"))
					pi.galleryType = GT_PAGE_WITH_THUMBS;
				else
					pi.galleryType = GT_ONE_PICTURE_PER_PAGE;
			}
			if (!strcmp(field, "Font")) {
				pi.font = value;
			}

			if (!strcmp(field, "count")) {
				int r = sscanf(value, "%d", &count);
				if (1 != r || count < 0) {
					fclose(f);
					return false;
				}
				thumbs = new RawImg[count];
				state = new bool[count];
				for (int i = 0; i < count; i++) {
					state[i] = true;
				}
			}
			if (!strncmp(field, "image[", 6)) {
				int index;
				int r = sscanf(field, "image[%d", &index);
				if (1 != r || index < 0 || index >= count) {
					fclose(f);
					return false;
				}
				bool paramfound = false;
				if (strstr(field, ".filename")) {
					thumbs[index].FileName = value;
					thumbs[index].outFileName = GetFn(value);
					paramfound = true;
				}
				if (strstr(field, ".outfilename")) {
					thumbs[index].outFileName = value;
					paramfound = true;
				}
				if (strstr(field, ".rotation")) {
					int v;
					int r = sscanf(value, "%d", &v);
					if (1 != r) {
						fclose(f);
						return false;
					}
					if (go)
						v = -1;
					thumbs[index].set_rotation(v);
					paramfound = true;
				} 
				if (strstr(field, ".enabled")) {
					int v;
					int r = sscanf(value, "%d", &v);
					if (1 != r) {
						fclose(f);
						return false;
					}
					state[index] = (v==0?false:true);
					paramfound = true;
				}
				if (strstr(field, ".resolution")) {
					int v;
					int r = sscanf(value, "%d", &v);
					if (1 != r || v < 0 || v > 99999) {
						fclose(f);
						return false;
					}
					thumbs[index].set_resolution(v);
					paramfound = true;
				}
				if (!paramfound) {
					fclose(f);
					return false;
				}
			}
		}
	}
	fclose(f);
	if (recreate_thumbs)
		return RecreateThumbs();
	return true;
}

bool Project:: SaveToFile(const wxString & fn)
{
	FILE *f;

	f = fopen(fn.c_str(), "wt");
	if (!f) return false;
	fprintf(f, "Title=%s\n", pi.Title.c_str());
	fprintf(f, "InputPath=%s\n", pi.iPath.c_str());
	fprintf(f, "OutputPath=%s\n", pi.oPath.c_str());
	fprintf(f, "GalleryType=%s\n", pi.galleryType == GT_PAGE_WITH_THUMBS ? 
		"PageWithThumbs" : "OnePicturePerPage");
	fprintf(f, "PicSize=%d\n", pi.PicSize);
	fprintf(f, "ThumbSize=%d\n", pi.ThumbSize);
	fprintf(f, "HasFrame=%d\n", pi.has_frame?1:0);
	///
	fprintf(f, "ThumbFormat=%s\n", (pi.ThumbFormat==FORMAT_JPEG)?"JPG":"PNG");
	fprintf(f, "BackgroundColor=#%06x\n", pi.BackgroundColor);
	fprintf(f, "TextColor=#%06x\n", pi.TextColor);
	fprintf(f, "Font=%s\n", pi.font.c_str());
	fprintf(f, "PicsPerRow=%d\n", pi.PicsPerRow);
	fprintf(f, "PicsPerGallery=%d\n", pi.PicsPerGallery);
	fprintf(f, "InfoPerImg=%d\n", pi.InfoPerImg);
	///
	fprintf(f, "BmpHandling=%s\n", BmpHandlingStr[pi.BmpHandling]);
	fprintf(f, "Rollover=%d\n", pi.Rollover);
	fprintf(f, "JpegQuality=%d\n", pi.JpegQuality);
	fprintf(f, "BackgroundImage=%s\n", pi.BackgroundImage.c_str());
	fprintf(f, "count=%d\n", count);

	for (int i = 0; i < count; i++) {
		fprintf(f, "image[%d].filename=%s\n", i, thumbs[i].FileName.c_str());
		fprintf(f, "image[%d].outfilename=%s\n", i, thumbs[i].outFileName.c_str());
		fprintf(f, "image[%d].rotation=%d\n", i, thumbs[i].get_rotation());
		fprintf(f, "image[%d].enabled=%d\n", i, state[i]?1:0);
		fprintf(f, "image[%d].resolution=%d\n", i, thumbs[i].get_resolution());
	}
	fclose(f);
	return true;
}

bool Project:: RecreateThumbs_implementation(bool master_thread)
{
	int i;
	progress_set_text(wxT("Recreating Thumbnails..."));
	progress_set_progress(0);
	for (i = counter.next(); i < count; i = counter.next()) {
		RawImg *thumb = thumbs+i;
		if (!thumb->AutoLoad()) {
			printf("Error While Loading `%s'!\n", thumb->FileName.c_str());
			return false;
		}
		int bigger_side = IMax(thumb->get_x(), thumb->get_y());
		if (bigger_side > pi.ThumbSize) {
			float ratio = (float) pi.ThumbSize / bigger_side;
			if (!thumb->Resize(
				ROUND(ratio*thumb->get_x()),
				ROUND(ratio*thumb->get_y()))
				) return false;
		}
		progress_set_progress((double) (i+1) / count);
		progress_set_text(wxString::Format(
			wxT("Recreating Thumbnails...%d of %d"), i+1, count), i+1);
	}
	return true;
}

bool Project:: NeedsRegeneration(RawImg& i, int bigger_side, bool isBmp, int forcedPicSize)
{
	if (!forcedPicSize)
		forcedPicSize = pi.PicSize;
	return (
		bigger_side > forcedPicSize ||
		i.get_rotation() ||
		(isBmp && pi.BmpHandling == BMP_HANDLING_TO_JPG)
		);
}

bool Project:: CreateGallery_init(void)
{
	globDirPath = SlashExtend(pi.oPath) + wxT("thumbs");
	if (!wxDirExists(globDirPath)) {
		if (!wxMkdir(globDirPath))
			return false;
	}
	if (pi.ThumbFormat == FORMAT_PNG) {
		static bool PNGHandler_added = false;
		if (!PNGHandler_added) {
			PNGHandler_added = true;
			wxPNGHandler *PNGHandler = new wxPNGHandler;
			wxImage::AddHandler(PNGHandler);
		}
	}
	
	progress_set_text(wxT("Creating Gallery..."));
	progress_set_progress(0);

	globDirPath += '/';
	return true;
}

bool Project:: CreateGallery_implementation(bool recreate_pics, bool master_thread)
{
	int i;
	int done = 0;
	int allCount = GetCount();
	for (i = counter.next(); i < count; i = counter.next()) {
		if (!state[i]) continue;
		done = doneCounter.next();
		bool isBmp =  ExtensionIs(thumbs[i].FileName, "bmp");
		wxString fn = globDirPath + (thumbs[i].outFileName);
		wxString fn_roll = globDirPath + "r_" + (thumbs[i].outFileName);
		wxString thumb = MakeExt(fn);
		wxString thumb_roll = MakeExt(fn_roll);
		
		float roll_opacity = (pi.Rollover == CHANGE_OPACITY ||
			pi.Rollover == HIGHLIGHT_AND_OPACITY) ? 0.80f : 1.00f;
		float roll_gamma = (pi.Rollover == HIGHLIGHT || 
			pi.Rollover == HIGHLIGHT_AND_OPACITY) ? 1.33f : 1.00f;
		if (pi.ThumbFormat == FORMAT_JPEG) {	
			thumbs[i].SaveToJPG(thumb.c_str(), pi.JpegQuality, pi.has_frame,
				pi.BackgroundColor, roll_opacity, 1.0f, DONT_RETAIN_EXIF);
			if (pi.Rollover)
				thumbs[i].SaveToJPG(thumb_roll.c_str(), pi.JpegQuality, pi.has_frame,
				pi.BackgroundColor, 1.0f, roll_gamma, DONT_RETAIN_EXIF);
		} else {
			thumbs[i].SaveToPNG(thumb.c_str(), pi.has_frame, roll_opacity);
			if (pi.Rollover)
				thumbs[i].SaveToPNG(thumb_roll.c_str(), pi.has_frame, 1.0f,
				roll_gamma);
		}
		progress_set_progress((done+0.5)/allCount);
		// check if we need to resize
		if (recreate_pics) {
			int bigger_side = IMax(thumbs[i].get_original_x(),
					thumbs[i].get_original_y());
			if (NeedsRegeneration(thumbs[i], bigger_side, isBmp, thumbs[i].get_resolution())) {
				RawImg big;
				big.LoadFromImageFile(thumbs[i].FileName.c_str());
				big.set_rotation(thumbs[i].get_rotation());

				int res_x, res_y;
				int picSize = thumbs[i].get_resolution();
				if (!picSize) picSize = pi.PicSize;
				if (CalculateResizedSizes(big, picSize, res_x, res_y)) {
					if (!big.Resize(res_x, res_y)) return false;
				}
				wxString big_fn = SlashExtend(pi.oPath) + (thumbs[i].outFileName);
				big.SaveToImageFile(MakeBMPH(big_fn).c_str(), pi.JpegQuality);
			} else {
				wxString big_fn = SlashExtend(pi.oPath) + (thumbs[i].outFileName);
				myCopyFile(thumbs[i].FileName.c_str(), big_fn.c_str());
			}
		}
		progress_set_progress((done+1)/ (double) allCount);
		progress_set_text(wxString::Format(
			wxT("Creating Gallery...%d of %d"), done+1, allCount), done+1);
	}
	if (master_thread)
		return CreateHtml();
	else
		return true;
}

wxString Project:: MakeExt(wxString ifn)
{
	wxString res = ifn.BeforeLast('.');
	if (pi.ThumbFormat == FORMAT_JPEG) {
		return res + ".jpg";
	} else {
		return res + ".png";
	}
}

wxString Project:: MakeBMPH(wxString ifn)
{
	wxString res = ifn.BeforeLast('.');
	if (pi.BmpHandling == BMP_HANDLING_TO_BMP && ExtensionIs(ifn, "bmp"))
		return res + ".bmp";
	else
		return res + ".jpg";
}



const ProjectIterator Project:: begin() 
{
	ProjectIterator i(-1, this);
	i++;
	return i;
}

const ProjectIterator Project:: end() 
{
	return ProjectIterator(count, this);
}


/*****************************************************************/
/** @class DoWorkThread                                          */
/*****************************************************************/

DoWorkThread:: DoWorkThread (Project * theProject, int what, SharedCounter *cnt)
: wxThread(wxTHREAD_DETACHED)
{
	prj = theProject;
	exec_fn = what;
	counter = cnt;
}

void* DoWorkThread:: Entry()
{
	bool master_thread = (0 != (exec_fn & 0x80000000));
	exec_fn &= ~0x80000000;
	int res;
	switch (exec_fn) {
		case TRD_RECREATE_THUMBS: 
			res = (int) prj->RecreateThumbs_implementation(master_thread);
			break;
		case TRD_GENERATE_GALLERY_TRUE:
			res = (int) prj->CreateGallery_implementation(true, master_thread);
			break;
		case TRD_GENERATE_GALLERY_FALSE:
			res = (int) prj->CreateGallery_implementation(false, master_thread);
			break;
		default: res = 0;
	}
	if (!counter->decrement())
		progress_end(res);
	return 0;
}

/*****************************************************************/
/** @class ProjectIterator                                       */
/*****************************************************************/

ProjectIterator:: ProjectIterator() 
{
	prj=NULL;
	current=-1;
}
ProjectIterator:: ProjectIterator(int seed, Project*p): 
	current(seed),prj(p) {}
void ProjectIterator:: operator = (const ProjectIterator & r)
{
	current = r.current;
	prj = r.prj;
}

bool ProjectIterator:: operator ==(const ProjectIterator & r)
{
	return current == r.current;
}
bool ProjectIterator:: operator !=(const ProjectIterator & r)
{
	return (current != r.current);
}

RawImg* ProjectIterator:: operator *()
{
	if (prj && current != -1)
		return prj->thumbs + current;
	else
		return NULL;
}

RawImg* ProjectIterator:: operator++(int)
{
	if (!prj) return NULL;
	int oc = current;
	current++;
	while (current < prj->count && !prj->state[current]) current ++;
	if (oc >=0 && oc < prj->count)
		return prj->thumbs + oc;
	else
		return NULL;
}

void ProjectIterator:: operator +=(int how_much)
{
	if (!prj) return;
	while (how_much--) {
		current ++;
		while (current < prj->count && !prj->state[current]) current++;
	}
}
