/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <wx/wx.h>
#include <wx/file.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

#include "project.h"
#include "bitmap.h"
#include "tools.h"
#include "defaults.h"

/** This is a dangerous define. We define what type of slashes delimit
    Filesystem structure. Under Win32 it's "\" and in UNIX & co. it's "/",
	but how to detect it?
*/
#ifdef _MSC_VER
#define OS_SLASH '\\'
#else
#define OS_SLASH '/'
#endif


int EndX(wxWindow * w)
{
	wxRect r = w->GetRect();
	return r.x + r.width;
}

int EndY(wxWindow * w)
{
	wxRect r = w->GetRect();
	return r.y + r.height;
}

int IMax(int a, int b)
{
	return (a>b?a:b);
}

wxString PathLastSlashStrip(wxString p)
{
	if (p == wxEmptyString) return p;
	if (p[p.length()-1]==OS_SLASH)
		p.Truncate(p.length()-1);
	if (-1 == p.Find(OS_SLASH))
		p += OS_SLASH;
	return p;
}

bool path_is_ok(const wxString & s)
{
	struct stat st;
	return (0 == stat(s.c_str(), &st));
}


bool parse_pair(const char * line, char * field, char * value)
{
	int i, j, l = (int) strlen(line);
	if (l < 2) return false;
	for (i = 0; i < l && '='!=line[i]; i++);
	if (i >= l-1) return false;
	
	for (j = 0; j < i; j++)
		field[j] = line[j];
	field[i] = 0;
	++i; j = 0;
	while (i < l)
		value[j++]=line[i++];
	if (value[j-1] == '\n') {
		value[j-1] = 0;
	} else {
		value[j] = 0;
	}
	return true;
}

bool ChooseFileName(Project * proj)
{
	wxString FN = wxFileSelector("Choose a filename to save your project",
		proj->Get_iPath(), "GalleryProject.agg", "*.agg", 
		"Anrieff Gallery Projects (*.agg)|*.agg",
		wxSAVE);
	if (FN.empty()) 
		return false;
	else {
		proj->FilenameToSave = FN;
		return true;
	}
}

unsigned BgColorRGB(const wxWindow *wnd)
{
	wxColor col = wnd->GetBackgroundColour();
	return ((col.Blue()) + (col.Green()<<8) + (col.Red() << 16));
}

wxString GetFn(const wxString & s)
{
	return s.AfterLast(OS_SLASH);
}

wxString SlashExtend(wxString src)
{
	if (src.empty()) return src;
	if (src[src.Length()-1] != OS_SLASH)
		return src + OS_SLASH;
	else
		return src;
}

wxString Capitalize(wxString src)
{
	for (unsigned i = 0; i < src.Length(); i++)
		if (src[i] >= 'a' && src[i] <= 'z') src[i] -= 32;
	return src;
}

bool ExtensionIs(wxString file, wxString ext)
{
	ext = Capitalize(ext);
	file = Capitalize(file);
	unsigned L1 = file.Length(), L2 = ext.Length();
	for (unsigned i = 1; i <= L2; i++)
		if (file[L1-i] != ext[L2-i]) return false;
	return true;
}

static inline int ToHex(char c)
{
	if (c >= '0' && c <= '9')
		return (int) (c - '0');
	if (c >= 'a' && c <= 'f')
		return (int) 10 + (c - 'a');
	return 0;
}

int GetHex(const wxString & s, int idx)
{
	return ToHex(s[idx])*16+ToHex(s[idx+1]);
}

unsigned Interpolate(unsigned color1, float where, unsigned color2)
{
	Rgb c1, c2;
	c1.CreateFrom(color1);
	c2.CreateFrom(color2);
	Rgb m;
	m.b = c1.b*(1.0f-where) + c2.b * where;
	m.g = c1.g*(1.0f-where) + c2.g * where;
	m.r = c1.r*(1.0f-where) + c2.r * where;
	return m.toRGB32();
}

unsigned GetFileSizeInKB(const wxString& fn)
{
	if (!wxFileExists(fn)) {
		//wxMessageBox("Can't get the size of file \"%s\"!", fn.c_str());
		return 0u;
	}
	wxFile temp(fn.c_str());
	return (unsigned) (temp.Length() / 1024u);
}

void scale_sizes(int & x, int & y, int bigger)
{
	int big = IMax(x, y);
	if (big > bigger) {
		float scale = (float)bigger/big;
		x = ROUND(scale*x);
		y = ROUND(scale*y);
	}
}

bool myCopyFile(const char *InputFile, const char *OutputFile)
{
	if (!strcmp(InputFile, OutputFile)) return true;
	FILE *fi, *fo;
	if (NULL == (fi = fopen(InputFile, "rb")))
		return false;
	if (NULL == (fo = fopen(OutputFile, "wb"))) {
		fclose(fi);
		return false;
	}
	unsigned char *buf[16384];
	int acread;
	while (16384 == (acread = fread(buf, 1, 16384, fi))) {
		fwrite(buf, 1, 16384, fo);
	}
	if (acread)
		fwrite(buf, 1, acread, fo);
	fclose(fi); fclose(fo);
	return true;
}

wxString stripExtension(wxString s)
{
	return s.BeforeLast('.');
}

Uint32 ByteSwap32(Uint32 x)
{
	x = ((x & 0xffff) << 16) | ((x >> 16) & 0xffff);
	x = ((x & 0x00ff00ff) << 8) | ((x >> 8) & 0x00ff00ff);
	return x;
}
Uint16 ByteSwap16(Uint16 x)
{
	x = ((x & 0xff) << 8) | ((x >> 8) & 0xff);
	return x;
}

// given an (possibly) unaligned pointer, return an 16-byte aligned one (possibly larger)
static inline void * perfectify(void *in)
{
	long t = reinterpret_cast<long>( in );
	t = (t&~15l)+16l;
	return reinterpret_cast<void*>( t );
}

// malloc's memory with forced 16-byte alignment
// Since malloc guarantees 8-byte alignment, we allocate 16 bytes more than
// we need. Then we perfectify() the pointer. We store the address of the
// original malloc() return just before the 16-byte boundary (that is,
// 4 or 8 bytes before the address, that is returned after)
void *sse_malloc(size_t size)
{
	void *res = malloc(size + 16);// malloc guarantees 8-byte alignment
	void *perfect = perfectify(res);
	void **store = (void**)perfect;
	*(--store) = res;
	
	return perfect;
}

void sse_free(void *ptr)
{
	void * to_free = *(((void**)ptr)-1);
	if (perfectify(to_free) == ptr) {
		free(to_free);
	} else {
		printf("Memory corruption!!!\n");
		printf("Something is wrong with sse_malloc/sse_free!\n");
	}
}

#ifdef __APPLE__
#define getCPUCount_defined
#include <unistd.h>
#include <mach/clock_types.h>
#include <mach/clock.h>
#include <mach/mach.h>

static int getCPUCount_system(void)
{
	kern_return_t kr;
	host_basic_info_data_t basic_info;
	host_info_t info = (host_info_t)&basic_info;
	host_flavor_t flavor = HOST_BASIC_INFO;
	mach_msg_type_number_t count = HOST_BASIC_INFO_COUNT;
	kr = host_info(mach_host_self(), flavor, info, &count);
	if (kr != KERN_SUCCESS) return 1;
	return basic_info.avail_cpus;
}
#endif

#ifdef _WIN32
#define getCPUCount_defined
#include <windows.h>
static int getCPUCount_system(void)
{
	SYSTEM_INFO system_info;
	GetSystemInfo(&system_info);
	return system_info.dwNumberOfProcessors;
}
#endif

#if defined linux || defined __linux__
#define getCPUCount_defined
#include <sys/sysinfo.h>
#include <unistd.h>
 
int getCPUCount_system(void)
{
	return sysconf(_SC_NPROCESSORS_ONLN);
}
#endif


int getCPUCount(void)
{
#ifdef getCPUCount_defined
	return getCPUCount_system();
#else
#	ifdef __GNUC__
#	warning "I don't know how to get the number of CPUs in your OS"
#	warning "Assuming 1 CPU - fix this if you are unhappy"
#	endif
	return 1;
#endif
}


void CalculateArrangement(wxWindow *wnd, wxPoint p[], wxSize s[], int count)
{
	wxRect rect = wnd->GetClientSize();
	int x = rect.width;
	int y = rect.height;
	//
	int required_width = count * DEFAULT_BUTTON_XSIZE + 10 * (count-1);
	int start_x = (x - required_width)/2;
	int start_y = y - DEFAULT_BUTTON_YSIZE - 15;
	for (int i = 0; i < count; i++) {
		p[i] = wxPoint(start_x, start_y);
		s[i] = wxSize(DEFAULT_BUTTON_XSIZE, DEFAULT_BUTTON_YSIZE);
		start_x += DEFAULT_BUTTON_XSIZE + 10;
	}
}

bool CalculateResizedSizes(const RawImg & img, int maxbig, int &x, int &y, bool rotated, bool original)
{
	int ix, iy; 
	if (rotated) {
		if (original) {
			img.get_original_rotated_sizes(ix, iy);
		} else {
			img.get_rotated_sizes(ix, iy);
		}
	} else {
		ix = original ? img.get_original_x() : img.get_x();
		iy = original ? img.get_original_y() : img.get_y();
	}
	int m = IMax(ix, iy);
	if (m <= maxbig) {
		x = ix;
		y = iy;
		return false;
	}		
	float ratio = (float) maxbig / m;
	x = ROUND(ix * ratio);
	y = ROUND(iy * ratio);
	return true;
}

/**
 * @class SharedCounter
*/

void SharedCounter::set(int value)
{
	lock.Lock();
	this->value = value;
	lock.Unlock();
}

void SharedCounter::reset()
{
	set(0);
}

int SharedCounter::next()
{
	int result;
	lock.Lock();
	result = (int) value++;
	lock.Unlock();
	return result;
}

int SharedCounter::decrement()
{
	int result;
	lock.Lock();
	result = (int) --value;
	lock.Unlock();
	return result;
}
