#include <math.h>
#include <string>
#include "chain.h"
using std::string;


Chain::Chain() 
{
	dim = -1;
}

Chain::Chain(Simplex s)
{
	dim = s.get_dim();
	s.norm();
	data.push_back(s);
	coeff.push_back((double)s.orientation());
}

Chain::Chain(const Complex & c, int d)
{
	for (int i = 0; i < c.size(); i++) {
		Simplex s = c[i];
		if (s.get_dim() != d)
			continue;
		s.norm();
		data.push_back(s);
		coeff.push_back((double) c[i].orientation());
	}
}

int Chain::size(void) const
{
	return (int)data.size();
}

/// Standard operations performed on chains:
/// +, -, *

Chain operator + (const Chain& a, const Chain& b)
{
	Chain c = a;

	for (int i = 0; i < b.size(); i++) {
		bool found = false;
		for (int j = 0; j < c.size(); j++)
			if (b.data[i] == c.data[j]) {
				found = true;
				c.coeff[j] += b.coeff[i];
				if (fabs(c.coeff[j]) < 1e-9) {
					c.coeff.erase(c.coeff.begin()+j);
					c.data.erase(c.data.begin()+j);
				}
				break;
			}
		if (!found) {
			c.data.push_back(b.data[i]);
			c.coeff.push_back((double) b.data[i].orientation());
		}
	}
	return c;
}

Chain operator - (const Chain& a, const Chain& b)
{
	Chain d = b;
	d *= -1.0;
	return a + d;
}


void Chain::operator *= (double scalar)
{
	for (int i = 0; i < data.size(); i++)
		coeff[i] *= scalar;
}

/// Create a complex from the current chain
Complex Chain::to_complex(void)
{
	int i;
	Complex c;
	for (i = 0; i < size(); i++)
		c.add(data[i]);
	
	int margin = c.size();
	i = 0;
	while (1) {
		for (i ; i < margin; i++) {
			Simplex s = c[i], ss;
			s.iter_start();
			while (s.iter(ss))
				c.add(ss);
		}

		if (margin == c.size())
			return c;

		i = margin;
		margin = c.size();
	}

	return c;
}

/// Input / output

std::istream& operator >> (istream& in, Chain& c)
{
	string s;
	in >> s;
	in >> s;
	int dim;
	sscanf(s.c_str(), "dim=\"%d\"", &dim);
	
	in >> s;
	in >> s;
	int N;
	sscanf(s.c_str(), "num=\"%d\"", &N);

	for (int i = 0; i < N; i++) {
		int k;
		cin >> k;

		Simplex ss;
		cin >> ss;

		if (ss.get_dim() != dim)
			continue;

		if (c.size() == 0)	{
			c.coeff.push_back(k);
			c.data.push_back(ss);
		} else {
			Chain cc(ss);
			cc *= k;
			c = c + cc;
		}
	}

	in >> s;
	in >> s;

	c.dim = dim;

	return in;
}

std::ostream& operator << (ostream& out, Chain& c)
{
	out << "<chain dim=\"" << c.dim << "\">" << endl;
	out << "<faces num=\"" << c.data.size() << "\"> ";

	for (int i = 0; i < c.data.size(); i++)
	{
		if (i)
			cout << " ";
		out << c.coeff[i] << " " << c.data[i];
	}

	out << " </faces>" << endl << "</chain>";

	return out;
}

/// Higher-level functions on chains

Chain accumulate(Chain start_val, std::vector<Chain> v, Chain (*fun)(const Chain& a, const Chain& b))
{
	Chain result = start_val;

	for (int i = 0; i < v.size(); i++)
		result = result + fun(result, v[i]);

	return result;
}

void map(std::vector<Chain> v, void (*fun) (Chain& a))
{
	for (int i = 0; i < v.size(); i++)
		fun(v[i]);
}

std::vector<Chain> filter(std::vector<Chain> v, bool (*fun) (const Chain& a))
{
	std::vector<Chain> result;

	for (int i = 0; i < v.size(); i++)
		if (fun(v[i]))
			result.push_back(v[i]);

	return result;
}
