#include "geometry.h"
#include "pointset.h"
#include "complex.h"
#include "abstract_sphere.h"
#include <string>
#include "vector3.h"
using std::string;
using std::endl;

Geometry::Geometry()
{
	points = NULL;
	data = NULL;
}

Geometry::Geometry(PointSet *p, Complex *c)
{
	points = p;
	data = c;
}

Geometry::~Geometry()
{
	if (points)
		delete points;
	if (data)
		delete data;
}

void Geometry::set_complex(Complex *c)
{
	data = c;
}

void Geometry::set_name(std::string n)
{
	name = n;
}

PointSet* Geometry::get_points() const
{
	return points;
}

Complex* Geometry::get_data() const
{
	return data;
}

/// Input / output

std::istream& operator >> (std::istream& in, Geometry& g)
{
	string s;
	in >> s;
	in >> s;
	s = s.substr(s.find_first_of("\"") + 1, s.size() - s.find_last_of("\"") + 1);
	g.set_name(s);

	PointSet *p = new PointSet();
	Complex *c = new Complex();

	in >> *p;
	in >> *c;

	g.points = p;
	g.data = c;

	return in;
}

std::ostream& operator << (std::ostream& out, Geometry& g)
{
	out << "\t\t<geometry name=\"" << g.name << "\">" << endl;
	out << *(g.points) << endl;
	out << *(g.data) << endl;
	out << "\t\t</geometry>" << endl;
	return out;
}

string Geometry::begin_jvx()
{
	return "<?xml version=\"1.0\" encoding=\"ISO-8859-1\" standalone=\"no\"?>\n<jvx-model>\n\t<geometries>\n";
}

string Geometry::end_jvx()
{
	return "\t</geometries>\n</jvx-model>\n";
}

/// Delaunay triangulation

/**
	Our approach is very simple and runs in O(N^4) for 2D and O(N^5) for 3D.
	
	Here is how we do the 2D Case:
	
	For each (noncolinear) three points, construct the circumcircle around them.
	(The construction is just determining where the circumcenter lies).
	
	If there are no other than the three points inside the circle, we add the
	three points and the triangle to the Delaunay triangulation.
	
	The 3D case is handled in a similar way, except for that we get 
	4 nonplanar points and construct circumsphere around them.
*/

bool Geometry::delaunay_triangulation(PointSet *ps, int dim)
{
	points = ps;
	//
	data = new Complex;

	// Which points are "used" for the construction of the current abstractsphere
	bool *used = new bool[points->size()];

	// Clear the points: no points are used in the beginning
	for (int i = 0; i < points->size(); i++)
		used[i] = false;
	//

	// the following recursion simulates 3 or 4 nested for loops (depending on dimension)
	recurse(dim, 0, used, -1);	
	return true; // the triangulation always succeeds
}

// params:
// dim - dimension (2 or 3)
// lev - how many points we have already chosen? (i.e. level of nesting in the for loops)
// used - specifies which point are already used
// last - specifies which point was used in the previous innermost for loop, so we iterate 
//        after it (purely an optimization, not something very important)
void Geometry::recurse(int dim, int lev, bool used[], int last)
{
	// if we created the number of points we need (3 in 2d, 4 in 3D)...
	if (lev == dim + 1) {
		AbstractSphere *sphere;
		if (dim == 3) sphere = new Sphere(); // If 3D, create a sphere, else
		else sphere = new Circle();          // just a circle

		Point input[4];
		int n = 0;
		for (int i = 0; i < points->size(); i++)
			if (used[i]) {
				input[n++] = points->operator [](i); // put the constraining points in the sphere/circle
			}

		if (sphere->construct(input)) { // construct the sphere/circle with the given points
			bool good = true;
			
			// for each point, which is not used in the construction...
			for (int i = 0; i < points->size(); i++)
				if (!used[i]) {
					if (sphere->is_inside(points->operator [](i))) { // check if it is inside the circle/sphere
						good = false;
						break;
					}
				}
			if (good) { // good. No points are inside the circle/sphere, add the triangle to the Delaunay triang.
				int pp[4];
				int n = 0;
				for (int i = 0; i < points->size(); i++)
					if (used[i]) pp[n++] = i; // create a list of the points, used in construction
					
				// construct a simplex using those points
				Simplex sim(dim, pp[0], pp[1], pp[2], pp[3]);
				
				// first, add low-dimension simplices
				for (int input_dim = 0; input_dim < dim; input_dim ++)
				{
					sim.iter_start(true);
					Simplex temp;
					while (sim.iter(temp)) {
						if (temp.get_dim() == input_dim)
							data->add(temp);
					}
				}
				
				// and last, add the big simplex (2D or 3D)
				data->add(sim);
			}
		}
		delete sphere;
		return;
	}
	for (int i = last + 1; i < points->size(); i++) { // here's the "for-loop" semantic
		used[i] = true;
		recurse(dim, lev + 1, used, i);
		used[i] = false;
	}
}
