#include "simplex.h"
#include <algorithm>
#include <string>
using std::string;

/// Returns the number of 1-bits in a number
static int popcount(int x)
{
	int c = 0;
	while (x) {
		c++;
		x &= x-1;
	}
	return c;
}


//////////////////////////////////////////////
/// @class Simplex
//////////////////////////////////////////////

Simplex::Simplex () {}

Simplex::Simplex (int kk, int a, int b, int c, int d)
{
	k = kk;
	verts[0] = a;
	verts[1] = b;
	verts[2] = c;
	verts[3] = d;
}

/// Init iterator
void Simplex::iter_start(bool all_sub_simplices)
{
	_iter = 1;
	_allsubsimp = all_sub_simplices;
}

/// Return next side of a simplex s
bool Simplex::iter(Simplex &s)
{
	int size = (1 << k) * 2 - 1;
	// Find a proper sub-simplex, if one exists
	while (!_allsubsimp && popcount(_iter) != k-1 && _iter < size) 
		_iter++;

	// out of range
	if (_iter >= size) return false;

	// dimension of the sub-simplex == count of 1 bits in _iter
	s.k = popcount(_iter) - 1; // because 0-simplex is a point
	int j = 0;
	for (int i = 0; i < k + 1; i++)
		if ((1 << i) & _iter)
			s.verts[j++] = verts[i];
	_iter++;
	return true;
}

/// Checks if a simlpex is a side of the current one
bool Simplex::is_side(const Simplex& x)
{
	iter_start();
	Simplex y;
	while (iter(y)) {
		if (x == y) return true;
	}
	return false;
}

/// Checks if a simplex is a sub-simplex of the current one
bool Simplex::is_sub_simplex(const Simplex &x)
{
	iter_start(true);
	Simplex y;
	while (iter(y)) {
		if (x == y) return true;
	}
	return false;
}

/// Equality operator
bool Simplex::operator == (const Simplex &r) const
{
	if (k != r.k) return false;
	Simplex x = *this;
	Simplex y = r;
	x.norm();
	y.norm();
	
	for (int i = 0; i <= k; i++)
		if (x.verts[i] != y.verts[i]) return false;

	return true;
}

/// Returns the orintation of a simplex = {1, -1}
int Simplex::orientation(void) const
{
	int inversions = 0;
	for (int i = 0; i <= k; i++)
		for (int j = i + 1; j <= k; j++)
			if (verts[i] > verts[j]) inversions ++;

	return ((inversions%2 == 0) ? + 1 : -1);
}

/// Normalizes a simplex
void Simplex::norm(void)
{
	std::sort(verts, verts+k+1); // because 0-simplex is a point
}

/// Checks if a simplex s and the current one share a common side
bool Simplex::is_adjacent(const Simplex& s)
{
	if (s == *this)
		return false;

	if (s.k == k)
	{
		bool result = false;
		for (int i = 0; i <= k; i++)
			for (int j = 0; j <= k; j++)
				if (s.verts[j] == verts[i]) {
					result = true;
					break;
				}
		return result;
	}

	Simplex ss;
	iter_start();
	while (iter(ss))
		if (ss.is_adjacent(s))
			return true;

	return false;
}

int Simplex::get_dim() const
{
	return k;
}


/// Input / output

std::istream& operator >> (istream &in, Simplex& r)
{
	r.k = -1;
	string s;
	bool ende = false;
	while (!ende) {
		in >> s;
		if ( s.substr(0, 3) == "<f>" )
			s = s.substr(3);
		if (s.find("</f>") != -1) {
			s = s.substr(0, s.find("</f>"));
			ende = true;
		}
		int x;
		if (1 == sscanf(s.c_str(), "%d", &x))
			r.verts[++r.k] = x;
	}

	return in;
}

std::ostream& operator << (ostream &out, Simplex& s)
{
	out << "<f>";
	for (int i = 0; i <= s.k; i++) {
		if (i) out << " ";
		out << s.verts[i];
	}
	out << "</f>";
	return out;
}
